#!/usr/bin/env python3
import sys, os, gzip, argparse, datetime

#
# Command line options
#
def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument('-g', '--gtf', required=True, help='TSEBRA Annotation GTF')
    p.add_argument('-c', '--cds', required=False, default=None, help='TSEBRA CDS FASTA')
    p.add_argument('-a', '--faa', required=False, default=None, help='TSEBRA Amino Acid FASTA')
    p.add_argument('-o', '--outd', required=False, default='.', help='Output Directory')
    p.add_argument('-i', '--ipr', required=False, default=None, help='InterProScan Output Table')
    p.add_argument('-m', '--homologs', required=False, default=None, help='Synolog Zebrafish Homologs')
    p.add_argument('-b', '--basename', required=False, default='tsebra_fixed', help='Basename for Files')
    p.add_argument('--keep-orig-transcript-ids', required=False, default=False, action='store_true', help='Keep the original TSEBRA transcript IDs in all files.')
    # Check input arguments
    args = p.parse_args()
    args.outd = args.outd.rstrip('/')
    if not os.path.exists(args.gtf):
        sys.exit(f"Error: '{args.gtf}' not found.")
    if not os.path.exists(args.outd):
        sys.exit(f"Error: '{args.outd}' not found.")
    if args.cds is not None:
        if not os.path.exists(args.cds):
            sys.exit(f"Error: '{args.cds}' not found.")
    if args.faa is not None:
        if not os.path.exists(args.faa):
            sys.exit(f"Error: '{args.faa}' not found.")
    if args.ipr is not None:
        if not os.path.exists(args.ipr):
            sys.exit(f"Error: '{args.ipr}' not found.")
    if args.homologs is not None:
        if not os.path.exists(args.homologs):
            sys.exit(f"Error: '{args.homologs}' not found.")
    return args

#
# Parse Annotations for Gene-Transcript Pairs
#
def gtf_get_gene_transcript_pairs(gtf_f, log):
    fh = None
    if gtf_f.endswith('.gz'):
        fh = gzip.open(gtf_f, 'rt')
    else:
        fh = open(gtf_f, 'r')
    cnt = 0
    gene_trans_pairs = dict()
    for i, line in enumerate(fh):
        if line.startswith('#'):
            continue
        cols = line.strip('\n').split('\t')
        assert len(cols) == 9
        if cols[2] not in ['exon', 'CDS']:
            continue
        cmmt = cols[8].split(';')
        gene_id = None
        transcript_id = None
        for field in cmmt:
            if len(field) == 0:
                continue
            field = field.rstrip(' ').lstrip(' ').split(' ')
            assert len(field) == 2, f'Line {i+1}'
            if field[0] == 'gene_id':
                gene_id = field[1].rstrip('\'\"').lstrip('\'\"')
            elif field[0] == 'transcript_id':
                transcript_id = field[1].rstrip('\'\"').lstrip('\'\"')
        if transcript_id is not None:
            if transcript_id not in gene_trans_pairs:
                gene_trans_pairs[transcript_id] = gene_id
                cnt += 1
    fh.close()
    print(f'Loaded {cnt:,} gene-transcript ID pairs from the annotation GTF.\n', flush=True, file=log)
    return gene_trans_pairs

#
# Get the Zebrafish gene names from the Synolog Homologs table
#
def get_zebrafish_homologs(homologs_f, log):
    fh = None
    if homologs_f.endswith('.gz'):
        fh = gzip.open(homologs_f, 'rt')
    else:
        fh = open(homologs_f, 'r')
    homologs = dict()
    row = 0
    id_idx = None
    name_idx = None
    for i, line in enumerate(fh):
        if line.startswith('#'):
            continue
        row+=1
        cols = line.strip('\n').split('\t')
        # In the first data row, determine if drer is orgA or orgB. This will define the specific columns to check for gene names.
        if row == 1:
            if cols[2] == 'drer.def.def':
                id_idx = 3
                name_idx = 9
            elif cols[1] == 'drer.def.def':
                id_idx = 8
                name_idx = 4
            else:
                sys.exit("Error: Zebrafish (drer.def.def) not in the homologs file.")
        gene_id = cols[id_idx]
        gene_name = cols[name_idx]
        homologs[gene_id] = gene_name
    print(f'Found names for {row:,} gene IDs.\n', flush=True, file=log)
    return homologs


#
# Extract functional annotation from InterProscan
#
def get_functional_annot(ipr_f, gene_trans_pairs, log):
    assert type(gene_trans_pairs) is dict
    fh = None
    if ipr_f.endswith('.gz'):
        fh = gzip.open(ipr_f, 'rt')
    else:
        fh = open(ipr_f, 'r')
    interpro = dict()
    cnt = 0
    for i, line in enumerate(fh):
        if line.startswith('#'):
            continue
        cols = line.strip('\n').split('\t')
        if cols[9] != 'T':
            continue
        transcript = cols[0]
        # Match this transcript ID to its corresponding gene ID
        gene_id  = gene_trans_pairs.get(transcript, None)
        if gene_id is None:
            continue
        description = cols[5].rstrip('.,;')
        go_terms = None
        if len(cols) > 13:
            go_terms = cols[13]
        if gene_id not in interpro:
            cnt+=1
            interpro.setdefault(gene_id, ([], []))
        # Append the descriptions
        if description not in interpro[gene_id][0] and description != '-':
            interpro[gene_id][0].append(description)
        # Append the GO terms
        if go_terms is not None and go_terms != '-':
            go_terms = go_terms.lstrip('|').rstrip('|').split('|')
            for go in go_terms:
                if go not in interpro[gene_id][1]:
                    interpro[gene_id][1].append(go)
    print(f'Found {i:,} functional annotations for a total of {cnt:,} transcripts.\n', flush=True, file=log)
    return interpro

#
# Parse the CDS fasta and fix the names
#
def fix_cds_fasta(cds_f, gene_trans_pairs, outd, basename, log, orig_name=False):
    assert type(gene_trans_pairs) is dict
    infh = None
    if cds_f.endswith('.gz'):
        infh = gzip.open(cds_f, 'rt')
    else:
        infh = open(cds_f, 'r')
    outf = f'{outd}/{basename}.cds.fa.gz'
    outfh = gzip.open(outf, 'wt')
    seqs  = 0
    fixed = 0
    for i, line in enumerate(infh):
        if line.startswith('#'):
            continue
        if line.startswith('>'):
            seqs += 1
            trans_id = line.strip('\n').lstrip('>')
            new_id = f'>{trans_id}\n'
            if orig_name is False:
                gene_id = gene_trans_pairs.get(trans_id, None)
                if gene_id is None:
                    sys.exit(f"Error: transcript ID ({trans_id}) in CDS not in GTF.")
                new_id = f'>{gene_id}\n'
                fixed += 1
            outfh.write(new_id)
        else:
            outfh.write(line)
    outfh.close()
    infh.close()
    print(f'Read {seqs:,} sequences from the CDS FASTA.\n    Added gene IDs to {fixed:,} transcripts.\n', flush=True, file=log)

#
# Parse the FAA fasta and fix the names
#
def fix_faa_fasta(faa_f, gene_trans_pairs, outd, basename, log, orig_name=False):
    assert type(gene_trans_pairs) is dict
    infh = None
    if faa_f.endswith('.gz'):
        infh = gzip.open(faa_f, 'rt')
    else:
        infh = open(faa_f, 'r')
    outf = f'{outd}/{basename}.protein.fa.gz'
    outfh = gzip.open(outf, 'wt')
    seqs  = 0
    fixed = 0
    for i, line in enumerate(infh):
        if line.startswith('#'):
            continue
        if line.startswith('>'):
            seqs += 1
            trans_id = line.strip('\n').lstrip('>')
            new_id = f'>{trans_id}\n'
            if orig_name is False:
                gene_id = gene_trans_pairs.get(trans_id, None)
                if gene_id is None:
                    sys.exit(f"Error: transcript ID ({trans_id}) in FAA not in GTF.")
                new_id = f'>{gene_id}\n'
                fixed += 1
            outfh.write(new_id)
        else:
            outfh.write(line)
    outfh.close()
    infh.close()
    print(f'Read {seqs:,} sequences from the FAA FASTA.\n    Added gene IDs to {fixed:,} proteins.\n', flush=True, file=log)

#
# Generate the new transcript IDs if needed
#
def get_new_transcript_ids(gene_trans_pairs, log):
    assert type(gene_trans_pairs) is dict
    new_gene_trans_pairs = dict()
    # Collapse the pairs by gene ID
    by_gene = dict()
    for trans in gene_trans_pairs:
        gene = gene_trans_pairs[trans]
        by_gene.setdefault(gene, [])
        by_gene[gene].append(trans)
    n_trans = 0
    # Tally the transcript in each gene
    for gene in by_gene:
        trans_list = by_gene[gene]
        t = 0
        for trans in sorted(trans_list):
            t+=1
            new_trans_id = f'{gene}.t{t}'
            new_gene_trans_pairs[trans] = new_trans_id
            n_trans+=1
    n_genes = len(by_gene)
    print(f'Reformatted {n_trans:,} transcript IDs for {n_genes:,} genes.\n', flush=True, file=log)
    return new_gene_trans_pairs

#
# Parse GTF and update annotations
#
def fix_annot_gtf(gtf_f, gene_trans_pairs, outd, basename, log, homologs=None, interpro=None, orig_names=False, date=datetime.datetime.now().strftime("%Y-%m-%d"), cmd=' '.join(sys.argv[:])):
    assert type(gene_trans_pairs) is dict
    assert type(orig_names) is bool
    infh = None
    if gtf_f.endswith('.gz'):
        infh = gzip.open(gtf_f, 'rt')
    else:
        infh = open(gtf_f, 'r')
    outf = f'{outd}/{basename}.gtf.gz'
    outfh = gzip.open(outf, 'wt')
    outfh.write(f'# {date}\n# {cmd}\n')
    seen_genes = list()
    cnt = 0
    genes = 0
    trans = 0
    for i, line in enumerate(infh):
        if line.startswith('#'):
            continue
        cols = line.strip('\n').split('\t')
        assert len(cols) == 9
        # Genes
        if cols[2] == 'gene':
            cnt += 1
            genes += 1
            def_cols = '\t'.join(cols[0:8])
            def_comt = cols[8].split(';, ')
            new_comt = list()
            if len(def_comt) != 1:
                sys.exit(f"Error: Line {i+1} in the file. Provide the default GTF from TSEBRA. Only the gene ID (g_XXXX) shold be present on the 9th column for `gene` rows.")
            # Fix the Gene ID
            gene_id = def_comt[0]
            if gene_id not in seen_genes:
                seen_genes.append(gene_id)
            # Fix the Gene Name
            gene_name = 'NA'
            if homologs is not None:
                gene_name = homologs.get(gene_id, 'NA')
            # Fix the Gene Description
            gene_desc = ['NA']
            if interpro is not None:
                gene_desc = interpro.get(gene_id, [['NA'], ['NA']])[0]
            gene_desc_str = ','.join(gene_desc)
            # Generate new comment line
            cmmt = f'gene_id \"{gene_id}\"; gene_name \"{gene_name}\"; gene_desc \"{gene_desc_str}\"'
            outfh.write(f'{def_cols}\t{cmmt};\n')
        # Transcripts
        elif cols[2] == 'transcript':
            cnt += 1
            trans += 1
            def_cols = '\t'.join(cols[0:8])
            old_tran = cols[8].rstrip('\"\'').lstrip('\"\'')
            new_tran = gene_trans_pairs.get(old_tran, None)
            if new_tran is None:
                sys.exit(f"Error: Line {i+1} of the file. Transcript {old_tran} does not have an associated gene id.")
            gene_id = new_tran.split('.')[0]
            if gene_id not in seen_genes:
                sys.exit(f"Error: Line {i+1} of the file. Transcript ({old_tran}) seen before its associated gene id ({gene_id}.")
            if orig_names is True:
                new_tran = old_tran
            cmmt = f'transcript_id \"{new_tran}\"; gene_id \"{gene_id}\"'
            outfh.write(f'{def_cols}\t{cmmt};\n')
        # Main gene annotations
        elif cols[2] in ['CDS','exon','start_codon','stop_codon']:
            cnt += 1
            def_cols = '\t'.join(cols[0:8])
            def_comt = cols[8].split(';')
            new_comt = list()
            gene_id = None
            old_tran = None
            for field in def_comt:
                if len(field) == 0:
                    continue
                field = field.rstrip(' ').lstrip(' ')
                if field.startswith('transcript_id'):
                    field = field.rstrip(' ').lstrip(' ').split(' ')
                    assert len(field) == 2, f'Line {i+1}'
                    old_tran = field[1].rstrip('\'\"').lstrip('\'\"')
                    new_tran = gene_trans_pairs.get(old_tran, None)
                    if new_tran is None:
                        sys.exit(f"Error: Line {i+1} of the file. Transcript {old_tran} does not have an associated gene id.")
                    gene_id = new_tran.split('.')[0]
                    if gene_id not in seen_genes:
                        sys.exit(f"Error: Line {i+1} of the file. Transcript ({old_tran}) seen before its associated gene id ({gene_id}.")
                    if orig_names is True:
                        new_tran = old_tran
                    tran_fld = f'transcript_id \"{new_tran}\";'
                    new_comt.append(tran_fld)
                elif field.startswith('gene_id'):
                    field = field.rstrip(' ').lstrip(' ').split(' ')
                    assert len(field) == 2, f'Line {i+1}'
                    field = field[1].rstrip('\"\'').lstrip('\"\'')
                    if field != gene_id:
                        sys.exit(f"Error: Line {i+1} of the file. Gene ID in this line ({field}) and does not match ID for annotation ({gene_id}).")
                    gene_fld = f'gene_id \"{field}\";'
                    new_comt.append(gene_fld)
            cmmt = ' '.join(new_comt)
            cmmt = cmmt.rstrip(';')
            outfh.write(f'{def_cols}\t{cmmt};\n')
        # if i>2000:
        #     break
    outfh.close()
    infh.close()
    print(f'''Read {i:,} lines from the GTF file.
    Reformatted annotation for {cnt:,} lines.
    {genes:,} genes.
    {trans:,} transcripts.\n''', flush=True, file=log)

def main():
    CMMD = ' '.join(sys.argv[:])
    DATE = datetime.datetime.now().strftime("%Y-%m-%d")
    TIME = datetime.datetime.now().strftime("%H:%M:%S")
    args = parse_args()
    log = open(f'{args.outd}/correct_tsebra_annots.log', 'w')
    print(f'{DATE} - {TIME}. Started correcting the TSEBRA annotation files.\n', flush=True, file=log)
    # Load the annotations
    gene_trans_pairs = gtf_get_gene_transcript_pairs(args.gtf, log)
    # Parse the IPR table to get functional annotation
    interpro = None
    if args.ipr is not None:
        interpro = get_functional_annot(args.ipr, gene_trans_pairs, log)
    # Parse the Homologs to get the zebrafish gene names
    homologs = None
    if args.homologs is not None:
        homologs = get_zebrafish_homologs(args.homologs, log)
    # Create the new transcript IDs if required
    if args.keep_orig_transcript_ids is False:
        gene_trans_pairs = get_new_transcript_ids(gene_trans_pairs, log)
    # Fix the CDS and FAA is present
    if args.cds is not None:
        # Parse the CDS FASTA and substitue
        fix_cds_fasta(args.cds, gene_trans_pairs, args.outd, args.basename, log, args.keep_orig_transcript_ids)
    if args.faa is not None:
        # Parse the FAA FASTA and substitue
        fix_faa_fasta(args.faa, gene_trans_pairs, args.outd, args.basename, log, args.keep_orig_transcript_ids)
    # Parse the GTF and fix the transcript ids
    fix_annot_gtf(args.gtf, gene_trans_pairs, args.outd, args.basename, log, homologs, interpro, args.keep_orig_transcript_ids, DATE, CMMD)
    # Finish the Log
    DATE = datetime.datetime.now().strftime("%Y-%m-%d")
    TIME = datetime.datetime.now().strftime("%H:%M:%S")
    print(f'{DATE} - {TIME} - Done!',flush=True, file=log)
    log.close()

# Run Code
if __name__ == '__main__':
    main()
