#!/usr/bin/env python3
import sys, os, argparse

#
# Command line options
#
def parse_args():
    p = argparse.ArgumentParser(prog='extract_synolog_homologs.py')
    p.add_argument('-r', '--reference-homologs', required=True, help='File with the reference IDs')
    p.add_argument('-s', '--synolog-homologs-tsv', required=True, help='Path to the XXX-YYY_homologs.tsv')
    p.add_argument('-o', '--out-dir', required=True, help='Output directory')
    p.add_argument('-i', '--ref-id', required=False, default='drer.def.def', help='ID of ref annotation in the homologs file')
    # Check input arguments
    args = p.parse_args()
    args.out_dir = args.out_dir.rstrip('/')
    if not os.path.exists(args.out_dir):
        sys.exit(f"Error: '{args.out_dir}' does not exist.")
    if not os.path.exists(args.reference_homologs):
        sys.exit(f"Error: '{args.reference_homologs}' does not exist.")
    if not os.path.exists(args.synolog_homologs_tsv):
        sys.exit(f"Error: '{args.synolog_homologs_tsv}' does not exist.")
    return args

class RefGene:
    def __init__(self, gid, name):
        self.id = gid
        self.name = name
    def __str__(self):
        return f'{self.id} {self.name}'

def parse_ref_homologs_d(tsv):
    ref_homologs = dict()
    with open(tsv, 'r') as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            fields = line.strip('\n').split('\t')
            if not len(fields) == 2:
                sys.exit('Error: Check reference IDs file.')
            gene = RefGene(fields[0], fields[1])
            ref_homologs[gene.id] = gene
    return ref_homologs

class TargetGene:
    def __init__(self, org, gid='--', name='--', chr='--', sta='--', end='--'):
        if len(name) == 0:
            name = 'NA'
        self.org = org
        self.gid = gid
        self.name = name
        self.chr = chr
        self.sta = sta
        self.end = end
    def __str__(self):
        row = f'{self.org}\t{self.gid}\t{self.name}\t{self.chr}\t{self.sta}\t{self.end}'
        return row

def parse_synolog_homolog_tsv(tsv, ref_homologs, outdir='./', ref_org = 'drer.def.def'):
    ref_set = set(ref_homologs.keys())
    target_homologs = dict()
    pair_name = tsv.split('/')[-1].split('_')[0]
    outfh = open(f'{outdir}/{pair_name}.extracted_homologs.tsv', 'w')
    outfh.write('#RefId\tRefName\tRefOrg\tQueryOrg\tGeneID\tGeneName\tChr\tStartBP\tEndBP\n')
    Org = None
    first_line = True
    # Index of the fields to extract - if Ref is OrgB
    org = 1
    gid = 3
    name = 4
    chr = 5
    sta = 6
    end = 7
    ref = 8
    # Parse the homologs file and extract the target genes
    with open(tsv, 'r') as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            fields = line.strip('\n').split('\t')
            if first_line:
                orgs = [fields[1], fields[2]]
                if ref_org not in orgs:
                    sys.exit(f'Error: {ref_org} not in {orgs}')
                # Change is Ref is OrgA
                if ref_org == fields[1]:
                    org = 2
                    gid = 8
                    name = 9
                    chr = 10
                    sta = 11
                    end = 12
                    ref = 3
                Org = fields[org]
                first_line = False
            # If not one of the focal genes, skip
            if fields[ref] not in ref_set:
                continue
            ref_gene = ref_homologs[fields[ref]]
            assert isinstance(ref_gene, RefGene)
            # GeneName can be an empty field
            gene_name = fields[name]
            if len(gene_name) == 0 or gene_name in ['NA', 'None', 'NaN', 'nan']:
                gene_name = '--'
            target_gene = TargetGene(fields[org], fields[gid], fields[name], fields[chr], fields[sta], fields[end])
            target_homologs[fields[ref]] = target_gene
    # Print the Output
    for ref_id in ref_homologs:
        ref_gene = ref_homologs[ref_id]
        assert isinstance(ref_gene, RefGene)
        target_gene = target_homologs.get(ref_id, None)
        if target_gene is None:
            target_gene = TargetGene(org=Org)
        row = f'{ref_gene.id}\t{ref_gene.name}\t{ref_org}\t{target_gene}\n'
        outfh.write(row)

def main():
    args = parse_args()
    ref_homologs = parse_ref_homologs_d(args.reference_homologs)
    parse_synolog_homolog_tsv(args.synolog_homologs_tsv, ref_homologs, args.out_dir, args.ref_id)

# Run Code
if __name__ == '__main__':
    main()

